/* 
   File: SSDE.h

   This file contains the code for the Spectral Graph Embedding Method
   via Sampled Spectral Distance Embedding (SSDE). Given a graph this 
   program returns the embedding of the nodes into Cartesian space 
   where the Euclidean distances between nodes approximates the 
   path distance in the original graph. 
   
   Authors:
   
   Eli Bocek-Rivele, Ali Civril, and Jonathan Purnell. 
   
 */

// Spectral Graph Embedding Method
// Eli Bocek-Rivele
// boceke@cs.rpi.edu
// 4/26/05
// Ali Civril
// Jon Purnell

#include <cstdio>
#include <cstdlib>
#include <csignal>
#include <assert.h>
#include <iostream>
#include <cstring>
#include <fstream>
#include <cmath>
#include <deque>
#include <map>
#include <time.h>
#include <vector>
#include <string>
#include "../lib/alglib/linalg.h"


#ifndef SSDE_H
#define SSDE_H


#define INFINITY_SSDE 100000000
#define SMALL 0.000000001

/*
  Enum: Type of Sampling nodes
  RANDOM_SAMPLING - node sampling is done randomly
  GREEDY_SAMPLING - node sampling is done by choosing the farthest node
 */
#define RANDOM_SAMPLING 0
#define GREEDY_SAMPLING 1

// #define NUM_OF_SVS 25		// how many singular values are to be used(inverted) related with INV_TOL also
//#define SV_PERCENTAGE 0.9	// same concept as the one above


int NUM_DIMS = 5;
int SAMPLE_SIZE = 25;
double SVD_TOL = 0.000000000001; // 1e-12
double PI_TOL = 0.0000001; // 1e-7
double REG_PARAM = 3;		// this is the power of the first singular values in regularization parameter
int SAMPLING = GREEDY_SAMPLING;

using namespace std;

/*
  Class: Node

  This class holds the information for a node in the graph.
  
  Format: Node Name File
  
  - First line of the file has the number of nodes 
  - Subsequent lines correspond to each node and contain 
  the name of the node and the number of neighboring node 

  Format: Graph File

  - First line of the file has the number of nodes
  - Subsequent lines correspond to each edge and contains
  the names of the two nodes followed by the edge weight
  - Note that this graph must be an undirected connected graph.
  The code doesn't not check for this, and if the graph is not
  connected the code will produce erroneous results. 

  Format: Coordinates File
  
  - First line of the file has the number of nodes
  - Subsequent lines correspond to each node and contains
  the coordinates of the embedded node
 */
class Node {

public: 

  //int index;			// the number assigned to this vertex
  string name;
  int Tindex;
  int lowlink;
  bool stacked; 

  int *neighbors;	// array to keep the indices of neighbors
  double *distances;       // array to keep the distances of neighbors
  int neighborSize;	// the index to keep the current size of the neighbor array
  int maximumSize;

  // for BFS
  double distance;	// distance to the root

/*
	Constructor: Node

		Default constructor 
*/
  Node() {
	this->name = "";
    this->neighbors = new int[5];
    this->distances = new double[5];
    this->neighborSize = 0;
    this->maximumSize = 5;
    this->distance = (double)INFINITY_SSDE;
    this->Tindex = INFINITY_SSDE;
    this->stacked = false;
  }

  /* 
     Constructor: Node
     
     Basic constructor (puts space for 5 neighbors)

     Parameters:

     name - string with the name of this node
   */
  Node(string name) {
    this->name = name;
    this->neighbors = new int[5];
    this->distances = new double[5];
    this->neighborSize = 0;
    this->maximumSize = 5;
    this->distance = (double)INFINITY_SSDE;
    this->Tindex = INFINITY_SSDE;
    this->stacked = false;
  }

  /*
    Constructor: Node
     
     Constructor that allocates memory for info on node's neighbors. 

     Parameters:

     name - string with the name of this node
     num_neighbors - integer number of neighbors to this node
   */
  Node(string name, int num_neighbors) {
    this->name = name;
    this->neighbors = new int[num_neighbors];
    this->distances = new double[num_neighbors];
    this->neighborSize = 0;
    this->maximumSize = num_neighbors;
    this->distance = (double)INFINITY_SSDE;
    this->Tindex = INFINITY_SSDE;
    this->stacked = false;
  }

  ~Node() { 
    delete neighbors;
    delete distances;
  }
	
  /*
    Function: insert

    Adds neighbor and edge weight. 

    Parameters:

    neighbor - index of neighboring node
    dist - edge weight (i.e. distance) between this node and the neighbor
   */
  void insert(int neighbor, double dist) {
    if(this->neighborSize == this->maximumSize)
      this->resize(2*this->maximumSize);

    neighbors[neighborSize] = neighbor;
    distances[neighborSize++] = dist;
  }

  /*
    Function: resize

    Allocates more memory for neighbor info.

    Parameters:
    
    newSize - new number of neighbors
   */
  void resize(int newSize) {
    int *oldArray = neighbors;
    double *oldDist = distances;
    neighbors = new int[newSize];
    distances = new double[newSize];

    for(int i = 0; i < maximumSize; i++ ) {
      neighbors[i] = oldArray[i];
      distances[i] = oldDist[i];
    }

    delete[] oldArray;
    delete[] oldDist;
	oldArray = NULL;
	oldDist = NULL;
    maximumSize = newSize;
  }
};

FILE *f = NULL;
FILE *g = NULL;
//double *xCoords, *yCoords;
vector< double* > Coords;
int numberOfNodes = 0;
Node **nodes = NULL;				// node array

int *samples = NULL;				// sampled vertices
double *minDistances = NULL;		
double **C = NULL;						// column sampling matrix
double **PHI = NULL;				// intersection of column and row sampling matrices
double **INV = NULL;				// inverse of PHI

int smallestDegreeNode;				// index of the smallest degree node
int firstSample;		// the first sampled node

map<string, int> node_map; // maps the node names to their index in nodes[]

/*
  Function: dotProduct
  
  Calculate dot product of two vectors. 
 */
//inline
double dotProduct(double a[], double b[]){
	
	double product = 0.0;
	if(a == NULL || b == NULL) return product;

	for(signed int i=1; i <= numberOfNodes; i++){
		product += a[i]*b[i];
	}
	return product;
}

//inline
double dotProduct(vector<double> &a, vector<double> &b) {
	double product = 0.0;
	if( (int)a.size() != (numberOfNodes+1) ||
		(int)b.size() != (numberOfNodes+1) ) return product;

	for(signed int i=1; i <= numberOfNodes; i++){
		product += a[i]*b[i];
	}
	return product;
}

/*
  Function: mag

  Calculate magnitude of vector.
*/
//inline
double mag(double a[]) {
	return sqrt(dotProduct(a, a));
}

//inline
double mag(vector<double> &a) {
	return sqrt(dotProduct(a, a));
}

/*
  Function: normalize

  Normalizes vector in-place
*/

//inline
double* normalize(double a[]) {
	
	if(a == NULL) { return a; }
	double magnitude = mag(a);

	for(signed int i=1; i <= numberOfNodes; i++){
		a[i] = a[i]/magnitude;
	}

	return a;
}

//inline
void normalize(vector<double> &a) {

	if( (int)a.size() != (numberOfNodes + 1)) return;

	double magnitude = mag(a);

	for(signed int i=1; i <= numberOfNodes; i++){
		a[i] = a[i]/magnitude;
	}

	//return a;
}

/*
  Function: multCons

  Returns vector resulting from multiplying input vector by input constant.
*/
//inline
double* multCons(double a[], double constant) {
	if(a == NULL) return NULL;

	double *b = new double[numberOfNodes+1];

	for(signed int i=1; i <= numberOfNodes; i++){
		b[i] = a[i]*constant;
	}

	return b;
}

/*
  Function: minusSSDE

  Returns the difference between two vectors.
*/
//inline
double* minusSSDE(double a[], double b[]) {
	if(a == NULL || b == NULL) return NULL;

	double *c = new double[numberOfNodes+1];

	for(signed int i=1; i <= numberOfNodes; i++){
		c[i] = a[i]-b[i];
	}

	return c;
}

/*
  Function: multGamma

  Returns vector who's elements average to zero. 
*/
//inline
double* multGamma(double a[]) {
	if(a == NULL) return NULL;

	double *b = new double[numberOfNodes+1];
	double sum = 0.0;

	for(signed int i=1; i <= numberOfNodes; i++){
		sum += a[i];
	}

	sum *= -1.0/(double)numberOfNodes;

	for(signed int i=1; i <= numberOfNodes; i++){
		b[i] = sum + a[i];
	}

	return b;
}

/*
  Function: makeOrtho

  Makes first input vector orthogonal to second input vector. 
 */
  //V2 = minusSSDE(V2, multCons(V1, (dotProduct(V1, V2))));
//inline
void makeOrtho(double a[], double b[]) { 
	if(a == NULL || b == NULL) return; 

  double* c = multCons(b, dotProduct(b,a));
  if(c == NULL) return; 
  
  for(signed int i=1; i <= numberOfNodes; i++) { a[i] -= c[i]; }

  delete [] c;
  c = NULL;
}

/*
  Function: multL

  Multiplies C*INV*R
 */
// multiply by C*INV*R
//inline
double* multL(double a[]) {
	if(a == NULL) return NULL;

	double *b = new double[SAMPLE_SIZE+1];	// result of first multiplication
	double *c = new double[SAMPLE_SIZE+1];	// result of second multiplication

	double sum = 0.0;

	// multiply by C^T
	for(signed int i = 1; i <= SAMPLE_SIZE; i++) {
		sum = 0.0;
		for(int j = 1; j <= numberOfNodes; j++) {
		  sum += -0.5*C[j][i]*a[j];			
		}
		b[i] = sum;
	}

	// multiply the result by INV
	for(int i = 1; i <= SAMPLE_SIZE; i++) {
		sum = 0.0;
		for(int j =	1; j <= SAMPLE_SIZE; j++) {
		  sum += INV[i][j]*b[j];			
		}
		c[i] = sum;
	}

	// multiply the result by C
	for(int i = 1; i <= numberOfNodes; i++) {
		sum = 0.0;
		for(int j =	1; j <= SAMPLE_SIZE; j++) {
		  sum += C[i][j]*c[j];
		}
		a[i] = sum;
	}

	// free memory
	delete [] b;
	delete [] c;
	b = c = NULL;

	return a;
}

/*
  Function: multGammaLGamma

  Simplifies the sequence of multiplying input vector by Gamma,
  by L and then by Gamma again. 
 */
//inline
double* multGammaLGamma(double a[]) {
	if(a == NULL) return NULL;

  double *b,*c;

  b = multGamma(a);
  b = multL(b);
  c = multGamma(b);
  
  delete [] b;
  b = NULL;
  
  return c;
}

/*
  Function: readNodes

  Loads the list of nodes names.  The format of this file is outlined in <Node Name File>.
  A hash table  is create so that the node names can be
  translated to their index in nodes[] in constant time.

*/
void readNodes(char *fileName) {
	if(fileName == NULL) { cerr << "Error: empty file name for readNodes" << endl; exit(EXIT_FAILURE);}
  ifstream fh(fileName);
  if(!fh.good()) { cerr << "Error: couldn't open " << fileName << endl; exit(EXIT_FAILURE); } 
  string nodeName;
  int num_neighbors;

  fh >> numberOfNodes;
  
  // prepare space for nodes
  nodes = new Node*[numberOfNodes+1];
  Coords.resize(NUM_DIMS,NULL);
  for(int dim = 0; dim < NUM_DIMS; dim++) {
    Coords[dim] = new double[numberOfNodes+1];
  }

  unsigned int sseed = (unsigned int)time(0);
  //sseed = 1294243397;
  cout << "Seed: " << sseed << endl;
  srand(sseed);

  for(int i = 1; i <= numberOfNodes; i++) {
    if( (i % 100000) == 0 ) { cout << (double)i/numberOfNodes << "% of data loaded\r"; } 
	if(!fh.good()) { cerr << "Error: insufficient data in nodes file: " << fileName << endl; exit(EXIT_FAILURE); }
    fh >> nodeName >> num_neighbors;
	if( (nodes[i] = new Node(nodeName,num_neighbors)) == NULL) {
		cerr << "Couldn't allocate space for nodes" << endl;
		exit(EXIT_FAILURE);
	}
   
    for(int dim = 0; dim < NUM_DIMS; dim++) {
      Coords[dim][i] = rand();
    }
    node_map[nodeName] = i;
  }
  cout << "100% of data loaded" << endl;

  cout << "Read " << numberOfNodes << " nodes" << endl;
}


/*
  Function: readFile 
  
  Loads the graph data. Format of this file follows <Graph File>.
*/
void readFile(char *fileName) {
	if(fileName == NULL) { cerr << "readFile received empty file name" << endl; exit(EXIT_FAILURE);}
  int count;
  string x,y;
  double w;
  ifstream edgeFile(fileName);
  if(!edgeFile.is_open()) { cerr << "Couldn't open file: " << fileName << endl; exit(EXIT_FAILURE);}
  
  count = 0;
  while( edgeFile >> x >> y >> w ) {

    // Insert them into the array also (will be used in average and scale)	  	 
    count++;

    if( node_map.find(x) == node_map.end() ) {
      cout << "Node " << x << " not found for edge with " << y << endl;
      continue;
    }
    if( node_map.find(y) == node_map.end() ) {
      cout << "Node " << y << " not found for edge with " << x << endl;
      continue;
    }
    nodes[ node_map[x] ]->insert( node_map[y],w);
    nodes[ node_map[y] ]->insert( node_map[x],w);

  }
  cout << endl;
  //fclose(f);
  edgeFile.close();
}

/*
  Function: while

  Writes the coordinates of the embedded nodes. Format of the file follows <Coordinates File>.
 */

void writeFile(char *fileName) {
	if(fileName == NULL) { cerr << "writeFile given empty file name" << endl; exit(EXIT_FAILURE);}
	ofstream fh(fileName);
	if(!fh.is_open()) { cerr << "writeFile: Couldn't open file " << fileName << endl; exit(EXIT_FAILURE); }
	fh << numberOfNodes << " " << NUM_DIMS << endl;
	if (strcmp(fileName, "final_coords.txt") == 0){
	  for(int i = 1; i <= numberOfNodes; i++) {
	  if(nodes[i]->neighborSize != 0) { // we ignore the nodes we didn't use
	    for(int dim = 0; dim < NUM_DIMS; dim++) {
	      fh << Coords[dim][i] / 10000. << " ";
	    }
	    fh << endl;
	  } else {
	    cout << "ignored node ";
	    cout << nodes[i]->name << endl;
	  }
	}
	}else{
	  for(int i = 1; i <= numberOfNodes; i++) {
	    if(nodes[i]->neighborSize != 0) { // we ignore the nodes we didn't use
	      for(int dim = 0; dim < NUM_DIMS; dim++) {
		fh << Coords[dim][i] << " ";
	      }
	      fh << endl;
	    } else {
	      cout << "ignored node ";
	      cout << nodes[i]->name << endl;
	    }
	  }
	}

	fh.close();
}

/*
  Function: BFS

  Breadth First Search. Calculates the distance from given node to all other nodes
  in the graph. 
 */
// takes the index of the node(pivot) to initiate BFS as a parameter (returns the furthest vertex)
int BFS(int pivot, int index, int samplingMethod) {
  
  std::deque<int> path;
	int front = 1;	   // the front of the path
	int back = 0;	   // the back of the path
	int current = 0;   // index of current node visited
	int neighbor_ind; // temp node to keep the bloody long pointer operation

	double max_dist = 0;
	int max_ind = 1;				// the furthest node to be returned
	int count = 0;
	
	assert(pivot <= numberOfNodes && pivot >= 1);
	path.push_front(pivot);
	nodes[pivot]->distance = 0;
	
	while( !path.empty() ) {
	  current = path.back(); path.pop_back();

	  for(int neighbor = 0; neighbor < nodes[current]->neighborSize; neighbor++) {
	    neighbor_ind = nodes[current]->neighbors[neighbor];
	    assert(neighbor_ind <= numberOfNodes);
	    if( (nodes[neighbor_ind]->distance == INFINITY_SSDE) || 
		(nodes[neighbor_ind]->distance > (nodes[current]->distance +
						  nodes[current]->distances[neighbor])) ){
	      
	      if( nodes[neighbor_ind]->distance == INFINITY_SSDE) { 		
		count++; 		
	      }

	      nodes[neighbor_ind]->distance = nodes[current]->distance +
		nodes[current]->distances[neighbor];
	      //path[++back] = nodes[neighbor_ind]->index;
	      path.push_front(neighbor_ind);
	    }	
	  }
		
	}

	max_dist = nodes[1]->distance;
	for(int i = 1; i <= numberOfNodes; i++) {
	  C[i][index] = nodes[i]->distance;
	  if( nodes[i]->distance > max_dist)
	    max_dist = nodes[i]->distance;
	}
	/*
	max_dist = nodes[1]->distance * nodes[1]->distance;
	for(int i = 1; i <= numberOfNodes; i++) {
	  C[i][index] = nodes[i]->distance *nodes[i]->distance;
	  if( nodes[i]->distance * nodes[i]->distance > max_dist)
		max_dist = nodes[i]->distance * nodes[i]->distance;
	}*/

	//cout << "farthest distance from this node: " << max_dist << endl;
	max_dist = 0;
	
	if(samplingMethod == GREEDY_SAMPLING) {
	  // updating distances
	  for(int i = 1; i <= numberOfNodes; i++) {
	    if(C[i][index] < minDistances[i]) 
	      minDistances[i] = C[i][index];
	  }
			
	  // find max distance
	  for(int i = 1; i <= numberOfNodes; i++) {
	    if(minDistances[i] > max_dist) {
	      max_dist = minDistances[i];
	      max_ind = i;
	    }
	  }
	  //cout << "Farthest node: Node " << max_ind << " at dist. " << max_dist << endl;
	}

	//delete[] path;
	//cout << "BFS nodes visited: " << count << endl;


	return max_ind;
}

void randomSampleCR();


/* 
   Function: allBFS

   Based on sampling method, samples nodes from the graph and determines
   the distances from the sampled nodes to all other nodes in the graph. 
   The C and PHI matrices are then computed. 

   Parameter:

   samplingMethod - indicates the sampling method

*/
void allBFS(int samplingMethod) {
  int sample;
  samples = new int[SAMPLE_SIZE+1];

  if(samplingMethod == 0) {

    randomSampleCR(); // randomly choose the nodes

    //call BFS to get distances of all nodes from each chosen node
    for(int i = 1; i <= SAMPLE_SIZE; i++) {
      for(int j = 1; j <= numberOfNodes; j++)
	nodes[j]->distance = INFINITY_SSDE;
      BFS(samples[i], i, samplingMethod);			// give the sampled vertex and its index
    }
  }
  // if we are doing clever sampling
  else {
    // randomly choose first node
    //srand(time(0));
    // sample = smallestDegreeNode;
    sample = (rand()%numberOfNodes)+1;
    firstSample = sample;

    // setup min distances of nodes from choosen nodes
    minDistances = new double[numberOfNodes+1];
    for(int i = 1; i <= numberOfNodes; i++)
      minDistances[i] = INFINITY_SSDE;

    // for each chosen node, find all distances and 
    // find farthest unchosen node (BFS return value)
    for(int i = 1; i <= SAMPLE_SIZE; i++) {
      //cout << i << ": finding distances for sample " << sample << endl;
      for(int j = 1; j <= numberOfNodes; j++)
	nodes[j]->distance = INFINITY_SSDE;
      // cout << sample << endl;
      samples[i] = sample;
      //cout << "Sample " << i << ": " << sample << endl;
      //cout << "calling bfs" << endl;
      sample = BFS(sample, i, samplingMethod);
    }
  }

  // form the squared distances for C
  cout << "forming C" << endl;
  for(int i = 1; i <= numberOfNodes; i++)
    for(int j = 1; j <= SAMPLE_SIZE; j++)
      C[i][j] *= C[i][j];

  // form PHI
  cout << "forming Phi" << endl;
  PHI = new double*[SAMPLE_SIZE+1];
  for(int i = 1; i < SAMPLE_SIZE+1; i++)
    PHI[i] = new double[SAMPLE_SIZE+1];

  cout << "forming Phi (b) " << endl;
  for(int i = 1; i <= SAMPLE_SIZE; i++) {
    for(int j = 1; j <= SAMPLE_SIZE; j++)
      PHI[i][j] = C[samples[j]][i];
  }	
}

/*
  Function: randomSampleCR

  Samples nodes from the graph randomly.
*/
// determine which columns(or rows) to be sampled
// result: the array 'samples' will hold the indices of the chosen columns
void randomSampleCR() {
  int sample;
  int count = 0;
  bool flag;
  //srand(time(0));
  while(count < SAMPLE_SIZE) {
    sample = (rand()%numberOfNodes) + 1;

    flag = false;
    for(int i = 1; i <= count; i++) {
      if(samples[i] == sample) {
	flag = true;
	break;
      }
    }

    if(!flag) 
      samples[++count] = sample;			// keep which columns are selected (used in the construction of PHI)
    else
      continue;
  }	
}

/*
  Function: house

  Returns the householder vector and the associated beta, given the size and the vector x.
 */
void house(double *v, double *x, int size, double *beta) {
  /* Golub-Van Loan Implementation */
  double sigma = 0.0;
  double mu;
  double temp;	// to keep the value of v[1]

  for(int i = 2; i <= size; i++)
    sigma += x[i]*x[i];

  v[1] = 1.0;
  for(int i = 2; i <= size; i++)
    v[i] = x[i];

  if(abs(sigma - 0.0) < SMALL)
    *beta = 0.0;
  else {
    mu = sqrt(x[1]*x[1] + sigma);
    if(x[1] <= 0)
      temp = v[1] = x[1] - mu;
    else
      temp = v[1] = -sigma/(x[1]+mu);

    *beta = 2*v[1]*v[1]/(sigma+v[1]*v[1]);
		
    for(int i = 1; i <= size; i++)
      v[i] = v[i]/temp;
  }
}

/*
  Function: bidiagonalize

  Bidiagonalizes a given matrix A by overwriting it, also returns Ue
  and Ve which are the elements in the expression Ue^T*A*Ve = B (A is
  overwritten as B) A is supposed to be an m*n matrix where m >= n, Ub
  is m*m, Vb is n*n
*/
void bidiagonalize(double **A, double **Ub, double **Vb, int m, int n) {
  double *v = new double[m+1];	// temporary householder vector
  double beta;					// and its associated beta
  double *temp = new double[m+1];	// temp vector

  // to accumulate Ue and Ve
  double *v1 = new double[m+1];
  double *v2 = new double[n+1];

  //DEBUGGING
  cout << "A(before):" << endl;
  for(int i = 1; i <= m; i++) {
	  for(int j = 1; j <= m; j++) 
		cout << A[i][j] << " ";
	  cout << endl;
  }

  // initialize Ub and Vb to Identity matrix
  for(int i = 1; i <= m; i++) 
    for(int j = 1; j <= m; j++) 
      Ub[i][j] = 0.0;
  for(int i = 1; i <= m; i++) 
    Ub[i][i] = 1.0;

  for(int i = 1; i <= n; i++) 
    for(int j = 1; j <= n; j++) 
      Vb[i][j] = 0.0;
  for(int i = 1; i <= n; i++) 
    Vb[i][i] = 1.0;

  // the main loop
  for(int j = 1; j <= n; j++) {
    // copy the related column of A to the temp vector (second line in the algorithm)
    for(int k = j; k <= m; k++)
      temp[k-j+1] = A[k][j];

    house(v, temp, m-j+1, &beta);

    // write over A by calculating A-beta*v*v^T*A(third line in the algorithm)
    // first keep the multiplication v^T*A in temp
    double total;
    for(int k = j; k <= n; k++) {
      total = 0.0;
      for(int l = j; l <= m; l++) 
	total += v[l-j+1]*A[l][k];
      temp[k-j+1] = total;
    }

    // then multiply it by v (on this pass, also calculate the final result)
    for(int k = j; k <= m; k++) 
      for(int l = j; l <= n; l++) 
	A[k][l] += -beta*v[k-j+1]*temp[l-j+1];	// A-beta...

    // restore the householder vector (fourth line in the algorithm)
    for(int k = j+1; k <= m; k++)
      A[k][j] = 0.0;

    // operations for calculating Ub (multiply by I-beta*v*v^T)
    // first, create the appropriate householder vector with leading zeros
    for(int k = 1; k <= j-1; k++)
      v1[k] = 0.0;

    for(int k = j; k <= m; k++)
      v1[k] = v[k-j+1];

    // multiply Ub with U_j (multiply by I-beta*v*v^T)
    // first keep the multiplication Ub*v1 in temp
    for(int k = 1; k <= m; k++) {
      total = 0.0;
      for(int l = 1; l <= m; l++) 
	total += v1[l]*Ub[k][l];
      temp[k] = total;
    }

    // then multiply it by v (on this pass, also calculate the final result)
    for(int k = 1; k <= m; k++) 
      for(int l = 1; l <= m; l++) 
	Ub[k][l] += -beta*temp[k]*v1[l];		// A-beta...

    if(j <= n-2) {
      // copy the related row of A to the temp vector (sixth line in the algorithm)
      for(int k = j+1; k <= n; k++)
	temp[k-j] = A[j][k];

      house(v, temp, n-j, &beta);

      // write over A by calculating A-beta*A*v*v^T(third line in the algorithm)
      // first keep the multiplication A*v in temp
      for(int k = j; k <= m; k++) {
	total = 0.0;
	for(int l = j+1; l <= n; l++) 
	  total += v[l-j]*A[k][l];
	temp[k-j+1] = total;
      }

      // then multiply it by v^T (on this pass, also calculate the final result)
      for(int k = j; k <= m; k++) 
	for(int l = j+1; l <= n; l++) 
	  A[k][l] += -beta*temp[k-j+1]*v[l-j];	// A-beta...

      // restore the householder vector (8th line in the algorithm)
      for(int k = j+2; k <= n; k++)
	A[j][k] = 0.0;

      // operations for calculating Vb (multiply by I-beta*v*v^T)
      // first, create the appropriate householder vector with leading zeros
      for(int k = 1; k <= j; k++)
	v2[k] = 0.0;

      for(int k = j+1; k <= n; k++)
	v2[k] = v[k-j];

      // first keep the multiplication Vb*v2 in temp
      for(int k = 1; k <= n; k++) {
	total = 0.0;
	for(int l = 1; l <= n; l++) 
	  total += v2[l]*Vb[k][l];
	temp[k] = total;
      }

      // then multiply it by v (on this pass, also calculate the final result)
      for(int k = 1; k <= n; k++) 
	for(int l = 1; l <= n; l++) 
	  Vb[k][l] += -beta*temp[k]*v2[l];		// A-beta...
    }
  }

  //DEBUGGING
  cout << "A(after):" << endl;
  for(int i = 1; i <= m; i++) {
	  for(int j = 1; j <= m; j++) 
		cout << A[i][j] << " ";
	  cout << endl;
  }
  cout << "Ub:" << endl;
  for(int i = 1; i <= m; i++) {
	  for(int j = 1; j <= m; j++) 
		cout << Ub[i][j] << " ";
	  cout << endl;
  }
  cout << "Vb:" << endl;
  for(int i = 1; i <= m; i++) {
	  for(int j = 1; j <= m; j++) 
		cout << Vb[i][j] << " ";
	  cout << endl;
  }
}

/*
  Function: givens
  
  computes cos(theta) c and sin(theta) c given a and b. 
  [c   s]^T[a] = [r].  [-s  c]^T[b] = [0].
*/
void givens(double a, double b, double *c, double *s) {
  double tau;
  if(b == 0) {
    *c = 1;
    *s = 0;
  }
  else {
    if(abs(b) > abs(a)) {
      tau = -a/b;
      *s = -1/sqrt(1+tau*tau);
      *c = *s*tau;
    }
    else {
      tau = -b/a;
      *c = 1/sqrt(1+tau*tau);
      *s = *c*tau;
    }
  }
}

/*
  Function: SVDstep

 Applies Givens rotations to the matrix on B bounded by upper and
 lower indices (call this Bbar). Overwrites Bbar with Ue^T*B*Ve, also
 calculates Ue and Ve in this formula Ue and Ve are treated as a whole
 (not only the parts remaining between upper and lower indices) They
 are first initialized to I, and accumulated accordingly.
*/
void SVDstep(double **B, double **Ue, double **Ve, int lower, int upper) {

  int n = upper-lower+1;
  int temp;		// for fast indexing
	
  double *d = new double[n+1];	// diagonal elements
  double *f = new double[n+1];		// superdiagonal elements

  // copy the diagonal and superdiagonal elements to the vectors defined above
  for(int i = 1; i <= n-1; i++) {
    temp = i+lower-1;
    d[i] = B[temp][temp];
    f[i] = B[temp][temp+1];
  }	
  d[n] = B[upper][upper];


  // find the eigenvalue mu which is closer to d[n]*d[n] + f[n-1]*f[n-1]
  double a = d[n-1]*d[n-1] + f[n-2]*f[n-2];
  double b = d[n]*d[n] + f[n-1]*f[n-1];
  double c = d[n-1]*f[n-1];

  double mu = (a+b+sqrt((a-b)*(a-b)+4*c*c))/2;
  double mu2 = (a+b-sqrt((a-b)*(a-b)+4*c*c))/2;

  // if mu2 is closer, pass the value to mu, (mu will be used in the proceeding steps)
  if(abs(mu2-b) < abs(mu-b))
    mu = mu2;
  
  double y = d[1]*d[1] - mu;
  double z = d[1]*f[1];

  double s;		// parameters of Givens rotations, c will also be used (defined above)
  double t1, t2, t3, t4;	// temporary variables to perform the matrix multiplicaton

  for(int k = 1; k <= n-1; k++) {
    // left transformation
    givens(y, z, &c, &s);

    t1 = f[k-1];
    t2 = d[k];
    t3 = f[k];
    t4 = d[k+1];

    f[k-1] = c*t1 - s*z;	// B[k-1][k]
    y = d[k] = c*t2 - s*t3;	// B[k][k]
    f[k] = s*t2 + c*t3;		// B[k][k+1]
    d[k+1] = c*t4;			// B[k+1][k+1]
    z = -s*t4;				// B[k+1][k]

    // postmultiply the result with Ve (look at Section 5.1.9 in Golub-Van Loan)
    for(int i = 1; i <= SAMPLE_SIZE; i++) {
      t1 = Ve[i][k];
      t2 = Ve[i][k+1];
		
      Ve[i][k] = c*t1 - s*t2;		
      Ve[i][k+1] = s*t1 + c*t2;	
    }

    // right transformation
    givens(y, z, &c, &s);

    t1 = d[k];
    t2 = f[k];
    t3 = d[k+1];
    t4 = f[k+1];

    d[k] = c*t1 - s*z;		// B[k][k]
    f[k] = c*t2 - s*t3;		// B[k][k+1]
    d[k+1] = s*t2 + c*t3;	// B[k+1][k+1]
    f[k+1] = c*t4;			// B[k+1][k+2]

    if(k < n-1) {
      y = f[k];	// B[k][k+1]
      z = -s*t4;	// B[k][k+2]
    }

    // postmultiply the result with Ve (look at Section 5.1.9 in Golub-Van Loan)
    for(int i = 1; i <= SAMPLE_SIZE; i++) {
      t1 = Ue[i][k];
      t2 = Ue[i][k+1];
		
      Ue[i][k] = c*t1 - s*t2;		
      Ue[i][k+1] = s*t1 + c*t2;	
    }
  }
  // copy the newly computed diagonal and superdiagonal elements back to B
  for(int i = 1; i <= n-1; i++) {
    temp = i+lower-1;
    B[temp][temp] = d[i];
    B[temp][temp+1] = f[i];
  }	
  B[upper][upper] = d[n];
}

/*
  Function: sample

  Calculates all the necessary matrices involved in sampling (also
  performs the SVD algorithm).  
*/
void sample() {
  cout << "Performing SVD" << endl;

  /*************** ALGLIB SVD ********************/
	// DEBUG ALGLIB
  alglib::real_1d_array a_w;
  alglib::real_2d_array a_A,a_u,a_vt;
  alglib::ae_int_t a_n = SAMPLE_SIZE;

  a_w.setlength(SAMPLE_SIZE);
  a_A.setlength(SAMPLE_SIZE,SAMPLE_SIZE);
  a_u.setlength(SAMPLE_SIZE,SAMPLE_SIZE);
  a_vt.setlength(SAMPLE_SIZE,SAMPLE_SIZE);
  
 for(int i = 0; i < SAMPLE_SIZE; i++) {
	 for(int j = 0; j < SAMPLE_SIZE; j++) {
		a_A[i][j] = PHI[i+1][j+1];		  
	 }
 }
  
 // svd
 if( alglib::rmatrixsvd(a_A,a_n,a_n,2,2,2,a_w,a_u,a_vt)) {
	cout << "alglib successfully converged" <<endl;
 }else cout << "alglib didn't converge" << endl;

// create inverse of diagonal matrix with singular values
 double max_s = a_w[0]*a_w[0]*a_w[0];
 alglib::real_2d_array a_Sig;
 a_Sig.setlength(SAMPLE_SIZE,SAMPLE_SIZE);
 for(int i = 0; i < SAMPLE_SIZE; i++) {
	 for(int j = 0; j < SAMPLE_SIZE; j++) {
		a_Sig[i][j] = 0.0;
		if(i == j) a_Sig[i][i] = a_w[i]/(a_w[i]*a_w[i] + max_s/(a_w[i]*a_w[i]));
	 }
 }

 // calculate INV = V*Sig*U^T
 alglib::real_2d_array a_INV;
 a_INV.setlength(SAMPLE_SIZE,SAMPLE_SIZE);
 alglib::rmatrixgemm(a_n,a_n,a_n, 1.0, 
	a_vt,0,0,1,
	a_Sig,0,0,0,
	0.0, a_INV,0,0); // V*Sig
 alglib::rmatrixgemm(a_n,a_n,a_n, 1.0,
	 a_INV, 0,0,0,
	 a_u,0,0,1,
	 0.0, a_INV,0,0);

  //form INV of PHI
  INV = new double*[SAMPLE_SIZE+1];
  for(int i = 1; i < SAMPLE_SIZE+1; i++)
    INV[i] = new double[SAMPLE_SIZE+1];

 for(int i = 0; i < SAMPLE_SIZE; i++) {
	 for(int j = 0; j < SAMPLE_SIZE; j++) {
		 INV[i+1][j+1] = a_INV[i][j];
	 }
 }

 cout << "Finished SVD" << endl;
 // // matrices involved in bidiagonalization
 // double **Ub = new double*[SAMPLE_SIZE+1];
 // double **Vb = new double*[SAMPLE_SIZE+1];

 // // matrices involved in SVD step
 // double **Ue = new double*[SAMPLE_SIZE+1];
 // double **Ve = new double*[SAMPLE_SIZE+1];

 // // matrices involved in final SVD step
 // double **U = new double*[SAMPLE_SIZE+1];
 // double **V = new double*[SAMPLE_SIZE+1];

 // for(int i = 0; i <= SAMPLE_SIZE; i++) {
 //   Ub[i] = new double[SAMPLE_SIZE+1];
 //   Vb[i] = new double[SAMPLE_SIZE+1];
 //   Ue[i] = new double[SAMPLE_SIZE+1];
 //   Ve[i] = new double[SAMPLE_SIZE+1];
 //   U[i] = new double[SAMPLE_SIZE+1];
 //   V[i] = new double[SAMPLE_SIZE+1];
 // }

 // cout << "Setup matrices" << endl;

 // // initialize Ue and Ve to Identity matrix (they are treated in full size)
 // for(int i = 0; i <= SAMPLE_SIZE; i++) 
 //   for(int j = 0; j <= SAMPLE_SIZE; j++) 
 //     Ue[i][j] = 0.0;
 // for(int i = 1; i <= SAMPLE_SIZE; i++) 
 //   Ue[i][i] = 1.0;

 // for(int i = 0; i <= SAMPLE_SIZE; i++) 
 //   for(int j = 0; j <= SAMPLE_SIZE; j++) 
 //     Ve[i][j] = 0.0;
 // for(int i = 1; i <= SAMPLE_SIZE; i++) 
 //   Ve[i][i] = 1.0;

	//cout << "initalized matrices" << endl;

	//

  /*******************Perform the Golub-Kahan SVD Algorithm*********************/
//
//  bidiagonalize(PHI, Ub, Vb, SAMPLE_SIZE, SAMPLE_SIZE);
//
//  cout << "bidiagonalized" << endl;
//  
//
//  int temp;		// to be used in the iteration to find p and q
//  int p, q;		// p and q as defined in Algorithm 8.6.2 (SVD Algorithm)
//  double total = 0.0;		// used for matrix multiplication (to accumulate U and V)
//  cout << "svding" << endl;
//  do {
//    for(int i = 1; i <= SAMPLE_SIZE-1; i++) {
//      if(abs(PHI[i][i+1]) <= SVD_TOL*(abs(PHI[i][i] + abs(PHI[i+1][i+1]))))
//	PHI[i][i+1] = 0.0;
//    }
//
//	// find q
//	temp = SAMPLE_SIZE;
//	q = 0;
//	while(temp > 1) {
//		if(PHI[temp-1][temp] == 0) {			
//			temp--;
//			q++;
//		}
//		else
//			break;
//	}
//
//	// find p
//	temp = 1;
//	p = 0;
//	while(temp < SAMPLE_SIZE-q) {
//		if(PHI[temp][temp+1] == 0) {
//			p = temp;
//		}
//		temp++;
//	}
//
//	bool flag = false;	// for checking whether we zeroed a superdiagonal element
//	if(q < SAMPLE_SIZE) {
//		for(int i = p+1; i <= SAMPLE_SIZE-q; i++) {
//			if(PHI[i][i] == 0) {
//				if(i < SAMPLE_SIZE)
//					PHI[i][i+1] = 0.0;		// note: out of bounds
//				flag = true;
//			}
//		}
//
//		if(!flag)
//			SVDstep(PHI, Ue, Ve, p+1, SAMPLE_SIZE-q);
//	}
//  }
//  while(q != SAMPLE_SIZE-1);
//
//  cout << "completed SVD" << endl;
//
//  // accumulate U by multiplying Ub and accumulated Ue's (U = Ub * Ue)
//  for(int i = 1; i <= SAMPLE_SIZE; i++) {
//	  for(int j = 1; j <= SAMPLE_SIZE; j++) {
//		  total = 0.0;
//		  for(int k = 1; k <= SAMPLE_SIZE; k++) {
//			  total += Ub[i][k]*Ue[k][j];
//		  }
//		  U[i][j] = total;
//	  }
//  }
//
//  // accumulate V by multiplying Vb and accumulated Ve's
//  for(int i = 1; i <= SAMPLE_SIZE; i++) {
//	  for(int j = 1; j <= SAMPLE_SIZE; j++) {
//		  total = 0.0;
//		  for(int k = 1; k <= SAMPLE_SIZE; k++)
//			  total += Vb[i][k]*Ve[k][j];
//		  V[i][j] = total;
//	  }
//  }
//cout << "finished accumlation" << endl;
//
//  // rearrange the singular values and the columns of U and V
//  int *indices = new int[SAMPLE_SIZE+1];		// array keeping the real indices of the singular values
//  for(int i = 1; i <= SAMPLE_SIZE; i++)
//    indices[i] = i;
//
//  // selection sort on the diagonal
//  int max;		// for keeping the index of the maximum element in one pass
//  double val;		// for swapping values
//  int index;		// for swapping indices
//  for(int i = 1; i < SAMPLE_SIZE; i++) {
//    max = i;
//    for(int j = i+1; j <= SAMPLE_SIZE; j++) {
//      if(abs(PHI[j][j]) > abs(PHI[max][max]))
//	max = j;
//    }
//    val = PHI[i][i];
//    PHI[i][i] = PHI[max][max];
//    PHI[max][max] = val;
//
//    index = indices[i];
//    indices[i] = indices[max];
//    indices[max] = index;
//  }
//
//	cout << "sorted values" << endl;
//
//  // rearrange the columns of U and V overwriting Ub and Vb(if there's a negative singular value, we negate the columns of U)
//	for(int j = 1; j <= SAMPLE_SIZE; j++) {
//		index = indices[j];
//		for(int i = 1; i <= SAMPLE_SIZE; i++) {
//			Ub[i][j] = U[i][index];
//		}
//		// if the entry is negative, negate th elements
//		if(PHI[j][j] < 0) {
//			for(int i = 1; i <= SAMPLE_SIZE; i++) {
//				Ub[i][j] *= -1;
//			}
//		}
//	}
//
//	// do the same thing for V
//	for(int j = 1; j <= SAMPLE_SIZE; j++) {
//		index = indices[j];
//		for(int i = 1; i <= SAMPLE_SIZE; i++) {
//			Vb[i][j] = V[i][index];
//		}
//	}
//
//  // finally make all the singular values positive on PHI
//  for(int i = 1; i <= SAMPLE_SIZE; i++) 
//    PHI[i][i] = abs(PHI[i][i]);
//
//  // form INV of PHI
//  INV = new double*[SAMPLE_SIZE+1];
//  for(int i = 1; i < SAMPLE_SIZE+1; i++)
//    INV[i] = new double[SAMPLE_SIZE+1];
//
//  double *inverses = new double[SAMPLE_SIZE+1];	// vector to keep the inverses of singular values
//
//  double alpha = pow(PHI[1][1], REG_PARAM);		// regularization constant (3 seems the best)
//  for(int i = 1; i <= SAMPLE_SIZE; i++)
//    inverses[i] = PHI[i][i]/(PHI[i][i]*PHI[i][i] + alpha/(PHI[i][i]*PHI[i][i]));	// regularized version
//
//  for(int i = 1; i <= SAMPLE_SIZE; i++) {
//	  for(int j = 1; j <= SAMPLE_SIZE; j++) {
//		  total = 0.0;
//		  for(int k = 1; k <= SAMPLE_SIZE; k++)
//			  total += Vb[i][k]*Ub[j][k]*inverses[k];
//		  INV[i][j] = total;
//	  }
//  }
//
//  
//
//  delete[] Ub;
//  delete[] Ue;
//  delete[] U;
//  delete[] V;
//  delete[] Vb;
//  delete[] Ve;
//
//  Ub = NULL;
//  Ue = NULL;
//  U = NULL;
//  V = NULL;
//  Vb = NULL;
//  Ve = NULL;
}

/*
  Function: powerIteration

  Performs power iterations to find the largest eigenvectors of 
  our distance matrix. 
 */
void powerIteration(){
  //cerr << "Power iteration..." << endl;
  /*****************************************
	/
	/			Power Iteration Method
	/
	/*****************************************/
  vector<double> theta(NUM_DIMS,0.0);		// two largest eigenvalues
  double temp;

  // initial guesses for eigenvectors
  vector< double* > Y(NUM_DIMS);
  vector< double* > V(NUM_DIMS);

  cout << "Finding largest eigenvalues by power iterations" << endl;
  //create the inital guess vectors for the power iteration
  cout << "init." << endl;
  //system("top");
  //srand(time(0));

  for(int dim = 0; dim < NUM_DIMS; dim++) {
    cout << "Dim. " << dim+1;
    // allocate space
    Y[dim] = new double[numberOfNodes+1];
    V[dim] = new double[numberOfNodes+1];
    // random init.
    cout << " init. ";
    for(int node=1; node < numberOfNodes+1; node++){
      Y[dim][node] = 2.0*(double)rand()/(double)RAND_MAX;
    }
    cout << "norm." << endl;
    normalize(Y[dim]);
  }

  double current, previous = SMALL;	// to keep track of the improvement ratio in power iteration

  int count, limit_count;
  limit_count = 0;

  for(int dim = 0; dim < NUM_DIMS; dim++) {
    count= 0; previous = SMALL;

    cout << "Working on dimension " << dim+1 << " of " << NUM_DIMS << "\r";
    while(true) {
      count++;
      // copy Y1 to V1
      for(int i = 1; i <= numberOfNodes; i++)
	V[dim][i] = Y[dim][i];

      // constrain
      for(int prev_dim = 0; prev_dim < dim; prev_dim++) {
	makeOrtho(V[dim],V[prev_dim]);
      }
      
      // update Y
      delete [] Y[dim]; // free up Y1 before setting 
      Y[dim] = multGammaLGamma(V[dim]);
      theta[dim] = dotProduct(V[dim], Y[dim]);
      normalize(Y[dim]);

      // find and test progress
      current = dotProduct(V[dim], Y[dim]);
      temp = abs(current/previous);
      //cout << "change: " << temp << endl;
      if((temp < 1+PI_TOL && temp > 1) || count > 100) {
	if(count > 100) limit_count++;
	break;
      }
      previous = current;
      //if( (count % 10) == 0 ) { cout << count << " " << temp << endl; }
    }

    cout << "Met iteration limit for " << limit_count << " of the " << NUM_DIMS << " dimensions." << endl;

    // finalize result
    for(int i = 1; i <= numberOfNodes; i++)
      V[dim][i] = Y[dim][i];

    //cout << "It took " << count << " iterations" << endl;
  }


  // theta now contains dominant eigenvalues
  // V now contains theta's eigenvectors

  //fclose(g);

  for(int dim = 0; dim < NUM_DIMS; dim++) {
    cout << "theta " << dim << " is: " << theta[dim] << endl;

    if(Coords[dim] != NULL) { delete[] Coords[dim]; }
    Coords[dim] = multCons(V[dim],sqrt(abs(theta[dim])));

    delete[] Y[dim];
    delete[] V[dim];
    Y[dim] = NULL;
    V[dim] = NULL;
  }
  cout << "Done with power iterations." << endl;
 
}

//void powerIteration(){
  //cerr << "Power iteration..." << endl;
  /*****************************************
	/
	/			Power Iteration Method
	/
	/*****************************************/
/*
  double theta1, theta2;		// two largest eigenvalues
  double temp;

  // initial guesses for eigenvectors
  double *Y1 = new double[numberOfNodes+1];
  double *Y2 = new double[numberOfNodes+1];

  double *V1 = new double[numberOfNodes+1];
  double *V2 = new double[numberOfNodes+1];



  //create the inital guess vectors for the power iteration
  srand(time(0));
  double random;
  for(int i=1; i < numberOfNodes+1; i++){
    random = (double)rand()/(double) RAND_MAX;
    Y1[i] = random + 2.12991777;		// just random numbers
    Y2[i] = random + 1.45345234;
  }

  normalize(Y1);
  normalize(Y2);

  double current, previous = SMALL;	// to keep track of the improvement ratio in power iteration


  int count = 0;
  cout << "First" << endl;
  while(true) {
    count++;
    // copy Y1 to V1
    for(int i = 1; i <= numberOfNodes; i++)
      V1[i] = Y1[i];

    delete [] Y1; // free up Y1 before setting 
    Y1 = multGammaLGamma(V1);
    theta1 = dotProduct(V1, Y1);
    normalize(Y1);

    current = dotProduct(V1, Y1);

    temp = abs(current/previous);
    if((temp < 1+PI_TOL && temp > 1) && count < 500) {
      break;
    }

    previous = current;

    if( (count % 10) == 0 ) { cout << count << " " << temp << endl; }
  }
  for(int i = 1; i <= numberOfNodes; i++)
    V1[i] = Y1[i];


  previous = SMALL;
	
  cout << count << endl;
  cout << "Second" << endl; // DEBUG
  count = 0; 
  while(true) {
    count++;
	  

    for(int i = 1; i <= numberOfNodes; i++)
      V2[i] = Y2[i];

    //V2 = minusSSDE(V2, multCons(V1, (dotProduct(V1, V2))));
    makeOrtho(V2,V1); // make V2 orthogonal to V2
    delete [] Y2; // free memory before setting Y2
    Y2 = multGammaLGamma(V2);
    theta2 = dotProduct(V2, Y2);
    normalize(Y2);

    current = dotProduct(V2, Y2);

    temp = abs(current/previous);
    if((temp < 1+PI_TOL && temp > 1)) {
      break;
    }

    previous = current;
    if( (count % 10) == 0 ) { cout << count << endl; }
  }
  cout << "Final count: " << count << endl;

  for(int i = 1; i <= numberOfNodes; i++)
    V2[i] = Y2[i];
  // theta1 now contains dominant eigenvalue
  // theta2 now contains the second-largest eigenvalue
  // V1 now contains theta1's eigenvector
  // V2 now contains theta2's eigenvector

  //fclose(g);

  cout << "theta1 is: " << theta1 << endl;
  cout << "theta2 is: " << theta2 << endl;
  //cout << "Done with power iteration." << endl;

  //populate the two vectors
  xCoords = multCons(V1, sqrt(abs(theta1)));
  yCoords = multCons(V2, sqrt(abs(theta2)));

  delete[] V1;
  delete[] V2;
  delete[] Y1;
  delete[] Y2;
  V1 = NULL;
  V2 = NULL;
  Y1 = NULL;
  Y2 = NULL;
  
  }*/

/*
	Function: calc_closeness_L

	Calculates the approx. of closeness centrality for each node
	by C*(INV*(C*1)), where 1 is column vector of ones. The matrices used
	are the approximation of L = D^2. 
*/
vector<double> calc_closeness_L() {
	vector<double> tmp_CC(SAMPLE_SIZE+1,0.0);
	vector<double> CC(numberOfNodes+1,0.0);

	// C*1
	for(int j = 1; j <= numberOfNodes; j++) {
		for(int i = 1; i <= SAMPLE_SIZE; i++) {		
			CC[i] += C[j][i];
		}
	}

	// INV*(C*1)
	for(int i = 1; i <= SAMPLE_SIZE; i++) {
		for(int j = 1; j <= SAMPLE_SIZE; j++) {
			tmp_CC[i] += INV[i][j] * CC[j];
		}
	}

	// C*(INV*(C*1))	
	for(int i = 1; i <= numberOfNodes; i++) {
		CC[i] = 0;
		for(int j = 1; j <= SAMPLE_SIZE; j++) {	
			CC[i] += tmp_CC[j]*C[i][j];
		}		
	}
	
	return CC;
}

/*
  Function: normalizeCoords

  Normalizes the coordinates of the embedded graph. 
 */
void normalizeCoords() {
  double min = INFINITY_SSDE, max = -INFINITY_SSDE;

  for(int dim = 0; dim < NUM_DIMS; dim++) {
    // find range
    for(int i = 1; i < numberOfNodes+1; i++) {
      if(min > Coords[dim][i])
	min = Coords[dim][i];
      else if(max < Coords[dim][i])
	max = Coords[dim][i];
    }
    
    // normalize 
    for(int i = 1; i < numberOfNodes+1; i++) {
      Coords[dim][i] = (Coords[dim][i] - min)*32768.0/(max-min);
    }
  }
}

/*
  Function: writeGraph
  
  Function for finding the Strongly Connected Components in the graph. 
  The largest SCC is written to coauthors_scc. 
*/
void writeGraph( std::deque<int> &S) {
  // re-index nodes 
  for(int i = 0; i < (int)S.size(); i++) {
    //nodes[S.front()]->index = i+1;
    nodes[S.front()]->stacked = false;
    S.push_back(S.front());
    S.pop_front();
  }

  // write graph
  
  int ind,indp;
  ofstream fh("coauthors_scc");
  fh << (int)S.size() << endl;
  while( ! S.empty() ) {
    ind = S.front(); S.pop_front();
    nodes[ind]->stacked = true;
    for(int i = 0; i < nodes[ind]->neighborSize; i++) {
      indp = nodes[ind]->neighbors[i];
      if( !nodes[indp]->stacked ) { // write edge
		  fh << nodes[ind]->name.c_str() << " " << nodes[indp]->name.c_str() 
			 << " " <<  nodes[ind]->distances[i] << endl;	
      }
    }
  }
  fh.close();
  
}

/*
  Function: Tarjan

  Function to perform the Tarjan algorithm.
*/
void Tarjan(int v, int &index, std::deque<int> &S, ofstream &scc_file) {
  int vp;
  nodes[v]->Tindex = index;
  nodes[v]->lowlink = index;
  
  // BFS - begin
  S.push_front(v);
  while(!S.empty()) {
    // print node
    v = S.back(); S.pop_back();
    scc_file << nodes[v]->name << "\t" << index << endl;

    // add unvisited nodes
    for(int i = 0; i < nodes[v]->neighborSize; i++) {
      vp = nodes[v]->neighbors[i];
      if(nodes[vp]->Tindex == INFINITY_SSDE) { // add to stack
	S.push_front(vp);
	nodes[vp]->Tindex = index;
      }
    }
  }

  // BFS - end
  
  index++;

  // Orig. Tarjan function
  /*
  //S.push_front(v);
  scc_file.open("sccs.txt",ios_base::app);
  scc_file << nodes[v]->name.c_str() << endl;
  scc_file.close();

  nodes[v]->stacked = true;
  for(int i = 0; i < nodes[v]->neighborSize; i++) {
    vp = nodes[v]->neighbors[i];
    if( nodes[vp]->stacked ) {
      nodes[v]->lowlink = (nodes[v]->lowlink < nodes[vp]->Tindex) ? nodes[v]->lowlink : nodes[vp]->Tindex;
    } else if( nodes[vp]->Tindex == INFINITY_SSDE ) {
      Tarjan(vp,index,S,scc_file);
      nodes[v]->lowlink = (nodes[v]->lowlink < nodes[vp]->lowlink) ? nodes[v]->lowlink : nodes[vp]->lowlink;
    }
  }

  if( nodes[v]->Tindex == nodes[v]->lowlink || index == 1000) {
    
    //if( S.size() > 50 ) {
    //  writeGraph(S);
    //} else {
    
    scc_file.open("sccs.txt",ios_base::app);
    scc_file << "###" << endl;
    scc_file.close();
    while(! S.empty()) { 
      nodes[ S.back() ]->stacked = false;
      S.pop_back(); 
    }
    exit(0);
    //}  
    
    }*/
}

void PrintCandInvPhi() {
  
  // output C
  ofstream cfile("C.txt");

  for(int i = 1; i <= numberOfNodes; i++) {
    for(int j = 1; j <= SAMPLE_SIZE; j++) {
      cfile << C[i][j] << " ";
    }
    cfile << endl;
  }

  cfile.close();

  // output Inverse Phi
  ofstream IPfile("InvPhi.txt");

  for(int i = 1; i <= SAMPLE_SIZE; i++) {
    for(int j = 1; j <= SAMPLE_SIZE; j++) {
      IPfile << INV[i][j] << " ";
    }
    IPfile << endl;
  }
  
  IPfile.close();
  
}

#endif 
